import os
import re
import unidecode
import importlib_resources
from .utils import create_key_and_request, check_password_match_to_key, if_file_exists, check_cert_match_key
from .utils import if_file_is_cert, create_pkcs12
from .cemaim import CeMaIm, get_mail_client_names, get_mail_client_shortcut
from PyQt5 import QtWidgets, uic, QtCore, QtGui
from pathlib import Path
#from PyQt5.QtCore import *
#from PyQt5.QtGui import *


class Gui:
    '''
    This class is for using GUI to
        - import certificates to email client
        - create request and key certificate
        - create pkcs12 certificate
    '''
    def __init__(self, config_map = None, debug = False):
        self.debug = debug
        self.app = QtWidgets.QApplication([])
        self.window = QtWidgets.QDialog()
        self.window.setMaximumWidth(521)
        self.window.setMaximumHeight(561)
        with importlib_resources.open_text('cemaim', 'cemaim.ui') as f:
            self.window = uic.loadUi(f)
#        with open('cemaim.ui', encoding='utf-8') as f:
#            uic.loadUi(f, self.window)
        self.config_map = config_map
        self.message = ""
        self.status_code = 0

        self.imp_user_cert_path = [Path("")]
        self.p12_user_cert_path = [Path("")]
        self.p12_user_key_path = [Path("")]

        self.__import_form__()
        self.__p12_form__()
        self.__create_request_form__()

        if self.config_map != None:
            self.__set_values_by_config_map__()
            self.__set_mail_client_by_config_map__()
        else:
            self.config_map = {}
            self.config_map['imp_mail'] = 'thb'

    def __set_mail_client_by_config_map__(self):
        '''
        Set comboBox of mail client by selected one from config file
        '''
        counter = 0
        for i in get_mail_client_names():
            if get_mail_client_shortcut(i) == self.config_map['imp_mail']:
                self.imp_mail.setCurrentIndex(counter)
                break
            counter += 1

    def __set_values_by_config_map__(self):
        '''
        Set values by what were readed from config file
        '''
        key_config_mapping  = {
            'req_country' : self.req_sub_C,
            'req_state_province' : self.req_sub_ST,
            'req_location' : self.req_sub_L,
            'req_organization' : self.req_sub_O,
            'req_folder' : self.req_location_path,
            'imp_ca' : self.imp_ca_path,
            'imp_folder' : self.imp_folder_path,
            'imp_user_cert' : self.imp_user_cert_butt,
            'p12_pem' : self.p12_user_cert,
            'p12_key' : self.p12_user_key
        }
        key_path_config_mapping = {
            'p12_pem' : self.p12_user_cert_path,
            'p12_key' : self.p12_user_key_path,
            'imp_user_cert' : self.imp_user_cert_path,
        }
        for (key,value) in self.config_map.items():
            if key in key_config_mapping.keys():
                if type(key_config_mapping[key]) == QtWidgets.QPushButton:
                    # Parse file
                    file_path = Path(value)
                    key_config_mapping[key].setText(file_path.name)
                else:
                    key_config_mapping[key].setText(value)
            if key in key_path_config_mapping.keys():
                key_path_config_mapping[key][0] = Path(value)
#                print(key, value)
#        self.req_sub_C.setText( self.config_map['req_country'])

    def errorMessage(self, text):
        '''
        Set warning error message when some error happen
        '''
        self.message = text
        self.status_code = 1
        if (self.debug == False):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle("Message box")
            msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
            msg.setText(text)
            retval = msg.exec()
        else:
            print(self.message)
            print(self.status_code)

    def get_file(self, folder_location_object, file_path, suffixes):
        '''
        File dialog menu and path will be stored at specified variable
        '''
        dialog = QtWidgets.QFileDialog(self.window, 'Open file', str(Path(file_path[0]).absolute()), suffixes)
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            file_path[0] = Path(dialog.selectedFiles()[0])
            folder_location_object.setText(Path(file_path[0]).name)

    def get_file_write_line(self, folder_location_object, suffixes):
        '''
        File dialog and path is stored to label
        '''
        dialog = QtWidgets.QFileDialog(self.window, 'Open file', str(Path(folder_location_object.text()).absolute()), suffixes)
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            file_path = Path(dialog.selectedFiles()[0])
            folder_location_object.setText(str(file_path))

    def get_folder(self, folder_location_object):
        '''
        File dialog to choose store folder
        '''
        dialog = QtWidgets.QFileDialog(self.window, 'Storage of certificate and key', folder_location_object.text())
        dialog.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            folder_location_object.setText(dialog.selectedFiles()[0])

    def __create_request_form__(self):
        '''
        Init request tab. Search button and set buttons methods
        '''
        self.req_name = self.window.findChild(QtWidgets.QLineEdit, 'req_name')
        self.req_surname = self.window.findChild(QtWidgets.QLineEdit, 'req_surname')
        self.req_location = self.window.findChild(QtWidgets.QPushButton, 'req_location')
        self.req_location_path = self.window.findChild(QtWidgets.QLabel, 'req_location_path')

        self.req_location.clicked.connect(lambda:self.get_folder(self.req_location_path))

        self.req_password = self.window.findChild(QtWidgets.QLineEdit, 'req_password')
        self.req_password_confirm = self.window.findChild(QtWidgets.QLineEdit, 'req_password_confirm')
        self.req_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.req_password_confirm.setEchoMode(QtWidgets.QLineEdit.Password)

        self.req_sub_C = self.window.findChild(QtWidgets.QLineEdit, 'req_sub_C')
        self.req_sub_ST = self.window.findChild(QtWidgets.QLineEdit, 'req_sub_ST')
        self.req_sub_L = self.window.findChild(QtWidgets.QLineEdit, 'req_sub_L')
        self.req_sub_O = self.window.findChild(QtWidgets.QLineEdit, 'req_sub_O')
        self.req_sub_EA = self.window.findChild(QtWidgets.QLineEdit, 'req_sub_EA')


        self.req_OK = self.window.findChild(QtWidgets.QPushButton, 'req_OK')
        self.req_OK.clicked.connect(lambda:self.preprocess_create_key_and_request())


    def preprocess_create_key_and_request(self):
        '''
        Do some check and run method to create key and request
        '''
        ## TODO check if password and confirm match
        name = unidecode.unidecode(self.req_name.text())
        name = re.sub('[ ]*$', '', name)
        name = re.sub('^[ ]*', '', name)
        surname = unidecode.unidecode(self.req_surname.text())
        surname = re.sub('[ ]*$', '', surname)
        surname = re.sub('^[ ]*', '', surname)

        if name == "" and surname == "":
            self.errorMessage("Name and surname are empty.")
            return
        if self.req_password.text() != self.req_password_confirm.text():
            self.errorMessage("Password and confirm password do not match.")
            return
        if self.req_password.text() == "" :
            self.errorMessage("Password is empty. Type some password.")
            return
        ## Check folder
        if os.path.exists(self.req_location_path.text()) == False:
            self.errorMessage("Path to this folder does not exist. Choose another one.")


        subject_dict = {}
        if self.req_sub_C.text() != "":
            subject_dict['C'] = self.req_sub_C.text()
        if self.req_sub_ST.text() != "":
            subject_dict['ST'] = self.req_sub_ST.text()
        if self.req_sub_L.text() != "":
            subject_dict['L'] = self.req_sub_L.text()
        if self.req_sub_O.text() != "":
            subject_dict['O'] = self.req_sub_O.text()
        if self.req_sub_EA.text() != "":
            subject_dict['emailAddress'] = self.req_sub_EA.text()

        ## TODO confirm message?

#        print(name, surname, self.req_password.text(), subject_dict, self.req_location_path.text())
        create_key_and_request(name, surname, self.req_password.text(), subject_dict, self.req_location_path.text())
        self.message = ""
        self.status_code = 0


    def __import_form__(self):
        '''
        Init import form for email client
        '''
        self.imp_mail = self.window.findChild(QtWidgets.QComboBox, 'imp_mail')
#        self.imp_ca_location = self.window.findChild(QtWidgets.QComboBox, 'imp_ca_location')
        self.imp_ca_path = self.window.findChild(QtWidgets.QLineEdit, 'imp_ca_path')
        self.imp_ca_butt = self.window.findChild(QtWidgets.QPushButton, 'imp_ca_butt')
#        self.imp_folder_location = self.window.findChild(QtWidgets.QComboBox, 'imp_folder_location')
        self.imp_folder_path = self.window.findChild(QtWidgets.QLineEdit, 'imp_folder_path')
        self.imp_folder_butt = self.window.findChild(QtWidgets.QPushButton, 'imp_folder_butt')
        self.imp_user_cert = self.window.findChild(QtWidgets.QLabel, 'imp_user_cert')
        self.imp_user_cert_butt = self.window.findChild(QtWidgets.QPushButton, 'imp_user_cert_butt')
        self.imp_password = self.window.findChild(QtWidgets.QLineEdit, 'imp_password')

        # Config Mail
        self.imp_mail.addItems(get_mail_client_names())
        self.imp_mail.currentIndexChanged.connect(lambda:self.mail_changed(self.imp_mail.currentText()))

        # Password
        self.imp_password.setEchoMode(QtWidgets.QLineEdit.Password)

        # Personal certificate
        self.imp_user_cert_butt.clicked.connect(lambda:self.get_file(self.imp_user_cert_butt, self.imp_user_cert_path, "Pkcs12 files (*.p12);;All files (*.*)"))
        self.imp_ca_butt.clicked.connect(lambda:self.get_file_write_line(self.imp_ca_path, "Pem files (*.pem);;All files (*.*)"))
        self.imp_folder_butt.clicked.connect(lambda:self.get_folder(self.imp_folder_path))

        self.imp_OK = self.window.findChild(QtWidgets.QPushButton, 'imp_OK')
        self.imp_OK.clicked.connect(lambda:self.preprocess_imp())

    def preprocess_imp(self):
        '''
        Run method to import certificates to mail client
        '''
        cemaim = CeMaIm(str(self.imp_ca_path.text()), str(self.imp_folder_path.text()), str(self.imp_user_cert_path[0]), self.config_map['imp_mail'], self.imp_password.text())
        status_code, message = cemaim.run()
        if status_code != 0:
                self.message = message
                self.errorMessage(message)
                return
        self.message = ""
        self.status_code = 0

    def __p12_form__(self):
        '''
        Init method for create pkcs12 certificate
        Initialize buttons and LineEdits
        '''
        self.p12_user_cert = self.window.findChild(QtWidgets.QPushButton, 'p12_user_cert')
        self.p12_user_key = self.window.findChild(QtWidgets.QPushButton, 'p12_user_key')
        self.p12_key_password = self.window.findChild(QtWidgets.QLineEdit, 'p12_key_password')
        self.p12_p12_password = self.window.findChild(QtWidgets.QLineEdit, 'p12_p12_password')
        self.p12_p12_password_confirm = self.window.findChild(QtWidgets.QLineEdit, 'p12_p12_password_confirm')

        self.p12_key_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.p12_p12_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.p12_p12_password_confirm.setEchoMode(QtWidgets.QLineEdit.Password)

        self.p12_user_cert.clicked.connect(lambda:self.get_file(self.p12_user_cert, self.p12_user_cert_path, "Pem files (*.pem);;All files (*.*)"))
        self.p12_user_key.clicked.connect(lambda:self.get_file(self.p12_user_key, self.p12_user_key_path, "Pem files (*.pem);;All files (*.*)"))

        self.p12_OK = self.window.findChild(QtWidgets.QPushButton, 'p12_OK')
        self.p12_OK.clicked.connect(lambda:self.preprocess_create_p12())

    def preprocess_create_p12(self):
        '''
        Run method to create pkcs12 certificate
        '''
        if if_file_exists(self.p12_user_cert_path[0]) == False:
            self.errorMessage(f'Personal certificate does not exist. There is nothing on this place:\n{self.p12_user_cert_path[0]}\n')
            return
        if if_file_exists(self.p12_user_key_path[0]) == False:
            self.errorMessage(f'Private key certificate does not exist. There is nothing on this place:\n{self.p12_user_key_path[0]}\n')
            return
        if if_file_is_cert(self.p12_user_cert_path[0]) == False:
            self.errorMessage(f'Personal certificate is not certificate. Choose another file')
            return
        if check_password_match_to_key(self.p12_user_key_path[0], self.p12_key_password.text()) == False:
            self.errorMessage("Password is incorrect for this private key certificate. Try another password or maybe you choosed bad private key certificate. Choosed file do not have to be private key.")
            return
        if check_cert_match_key(self.p12_user_cert_path[0], self.p12_user_key_path[0], self.p12_key_password.text()) == False:
            self.errorMessage("Personal certificate and private key certificate does not match. Choose different files.")
            return
        if self.p12_p12_password.text() != self.p12_p12_password_confirm.text():
            self.errorMessage("Password and confirm password do not match.")
            return
        if self.p12_p12_password.text() == "" :
            self.errorMessage("Password is empty. Type some password.")
            return

        output_folder = Path(self.p12_user_cert_path[0]).parent
#        print(self.p12_user_cert_path[0], self.p12_user_key_path[0], output_folder, self.p12_key_password.text(), self.p12_p12_password.text())
        create_pkcs12(self.p12_user_cert_path[0], self.p12_user_key_path[0], output_folder, self.p12_key_password.text(), self.p12_p12_password.text())

        self.message = ""
        self.status_code = 0

    @QtCore.pyqtSlot(str)
    def mail_changed(self, client_name):
        '''
        Change mail shortcut when mail client is changed
        '''
        self.config_map['imp_mail'] = get_mail_client_shortcut(client_name)

    def run(self):
        '''
        Run method for create and exit Gui
        '''
        self.window.show()
        return self.app.exec()

def run_gui():
    gui = Gui()
    gui.run()



import os
import sys
import re
import requests
import tempfile
from bs4 import BeautifulSoup
from os import listdir
from os.path import isfile, join
from .utils import *

## Import mail client file here
from .thb_certdb import Thunderbird, MailClient

# Place to add another email client communicator
mail_clients = {
    'thb': Thunderbird,
#    'mail': MailClient,
}

def cli_mail_client_text():
    '''
    Get text and list of mail clients
    This option is used by click
    '''

    mail_client_list = []
    output = "["
    for i in mail_clients.items():
        line = f'{i[0]} - ({i[1].mail_client_name})'
        output = output + line + ", "
        mail_client_list.append(line)
    output = re.sub(', $', '', output)
    output = output + "]"
    return output, mail_client_list

def get_mail_client_shortcut(client_name):
    '''
    Get shortcuts of clients names
    '''
    for (key, value) in mail_clients.items():
        if value.mail_client_name == client_name:
           return key

def get_mail_client_names():
    '''
    Return list of client names
    '''
    mail_client_names = []
    for i in mail_clients.items():
        mail_client_names.append(i[1].mail_client_name)
    return mail_client_names

class CeMaIm:
    '''
    Class for importing certificates to mail client.
    '''
    def __init__(self, auth_cert, path, user_cert, mail, password, debug = False, ca_location = False, certs_location = False):
        ca = re.sub('[ ]*$', '', auth_cert)
        ca = re.sub('^[ ]*', '', ca)

        p12 = re.sub('[ ]*$', '', user_cert)
        p12 = re.sub('^[ ]*', '', p12)
        self.auth_cert = ca
        self.path = path
        self.user_cert = p12
        self.mail = mail
        self.password = password
        self.ca_location = ca_location
        self.certs_location = certs_location
        self.debug = debug
        self.communicator = self.__get_communicator__()

    def __get_communicator__(self):
        '''
        Set up what mail client communicator will be used
        '''
        return mail_clients[self.mail](self.debug)

    def recognize_web(self, path):
        '''
        Recognize where the file is located if its on web or at local disk
        '''
        if re.match('^https?://.*', path):
#            print("Web")
            return True
        else:
#            print("Local")
            return False

    def listFD(self, url, ext='pem'):
        '''
            Get list of files at specified http site
        '''
        try:
            page = requests.get(url)
        except:
            return 30, f'Something goes wrong when downloading from "{self.path}".\nTry files download manual and set their path at disk.', []
        if page.status_code < 200 or page.status_code > 300:
            return 30, f'Something goes wrong when downloading from "{self.path}".\nTry files download manual and set their path at disk.', []
#        print(page)
        soup = BeautifulSoup(page.text, 'html.parser')
        return 0, '', [url + node.get('href') for node in soup.find_all('a') if node.get('href').endswith(ext)]


    def get_and_import_CA(self):
        '''
        Download and test file of certificate authority.

        If it goes through test it will be imported by mail client
        '''
        ca_file = ""

        fp = tempfile.NamedTemporaryFile()
#        print(fp.name)
        #Recognize if it is web or local
        if self.recognize_web(self.auth_cert) == True:
            #Download from web
            try:
                myfile = requests.get(self.auth_cert)
            except:
                return 20, f'Something goes wrong when downloading from "{self.auth_cert}".\nTry it again or download it manual and set its path at disk.'
            if myfile.status_code < 200 or myfile.status_code > 300:
                return 20, f'Something goes wrong when downloading from "{self.auth_cert}".\nTry it again or download it manual and set its path at disk.'
            ca_file = myfile.content.decode()
            fp.write(myfile.content)
#            fp.seek(0)
#            print(fp.read())
            ca_file = fp.name
#            print(myfile.content.decode())
        else:
            ca_file = self.auth_cert
#            with open(self.auth_cert) as f:
#                ca_file = f.read()

        # Import CA
#        self.communicator.import_ca("/mnt/Data/School/Master/PYT/certs-mail-import/CA/mipytCA/cacert.pem" , "tester")
        if if_file_exists(ca_file) == False:
            return 21, f'File \'{ca_file}\' you are trying import does not exist. Select another file.'
        if if_file_is_cert(ca_file) == False:
            return 22, f'File \'{ca_file}\' you are trying import is not certicate file at pem format. Select another file.'
        self.communicator.import_ca(ca_file, str(Path(self.auth_cert).name))
#        self.communicator.import_ca(ca_file, str(Path(self.auth_cert).name)self.auth_cert.split('/')[-1])
        fp.close()
        return 0, ''

    def get_and_import_pem(self):
        '''
        Download all from web or read all from specified path and imported by mail client
        Every file is test if exists and if its pem file.
        '''
        users_certs_list = []
        if self.recognize_web(self.path) == True:
            # Get list of users from web
            status_code, message, list_of_files = self.listFD(self.path, 'pem')
            if status_code != 0:
                return status_code, message
            for pem in list_of_files:
                fp = tempfile.NamedTemporaryFile()
                try:
                    myfile = requests.get(pem)
                except:
                    debug_message(self.debug, f'Something goes wrong when downloading from "{self.auth_cert}".\nTry it again or download it manual and set its path at disk.')
                    continue
                if myfile.status_code < 200 or myfile.status_code > 300:
                    debug_message(self.debug, f'Something goes wrong when downloading from "{self.auth_cert}".\nTry it again or download it manual and set its path at disk.')
                    continue

                fp.write(myfile.content)

                if if_file_exists(fp.name) == False:
                    debug_message(self.debug, f'File \'{ca_file}\' you are trying import does not exist. Will not be imported.')
                    fp.close()
                    continue
                if if_file_is_cert(fp.name) == False:
                    debug_message(self.debug, f'File \'{ca_file}\' you are trying import is not certicate file at pem format. Will not be imported.')
                    fp.close()
                    continue

                self.communicator.import_pem(fp.name, pem.split('/')[-1])
                fp.close()
        else:
            if os.path.exists(self.path) == False:
                return 31, f'Path to this folder {self.path} does not exist. Choose another one.'

            users_certs_list = [f for f in listdir(self.path) if f.endswith('.pem')]
            file_path = Path(self.path)
            for pem in users_certs_list:
                file_path_with_cert = file_path / pem
                if if_file_exists(str(file_path_with_cert)) == False:
                    debug_message(self.debug, f'File \'{str(file_path_with_cert)}\' you are trying import does not exist. Will not be imported.')
                    continue
                if if_file_is_cert(str(file_path_with_cert)) == False:
                    debug_message(self.debug, f'File \'{str(file_path_with_cert)}\' you are trying import is not certicate file at pem format. Will not be imported.')
                    continue

                self.communicator.import_pem(str(file_path_with_cert), str(file_path_with_cert.name))
        return 0, ''

    def import_personal_cert(self):
        '''
        Import personal cert at pkcs12 format
        '''
        self.communicator.import_personal_cert(self.user_cert, self.password)

    def check_p12_file(self):
        '''
        Check if pkcs12 file is OK and password match to cert
        '''
        if if_file_exists(self.user_cert) == False:
            return 10, f'Personal certificate does not exist. There is nothing on this place:\n{self.user_cert}\n'
        try:
            p12 = crypto.load_pkcs12(open(self.user_cert, 'rb').read(), self.password)
        except:
            return 11, f'Password is incorrect for this p12 certificate "{Path(self.user_cert).name}". Try another password or maybe you choosed bad private key certificate. Choosed file do not have to be private key.'
        return 0, ''

    def run(self):
        '''
        Main method that run importing of CA, pkcs12 and pems file
        '''
        if self.debug == True:
            print(f'Running with configuration:\n\tCA: {self.auth_cert}\n\tMail client: {self.mail}\n\tUser certificate: {self.user_cert}\n\tPassword: {self.password}\n\tPath: {self.path}\n')
        # Test if mail client is OK
        if self.communicator.lock != 0:
            return self.communicator.lock, self.communicator.lock_message


        ## Get and import CA
        if self.auth_cert != "":
            status_code, message = self.get_and_import_CA()
            if status_code != 0:
                return status_code, f'{message}\n If you want to skip import CA set path to certificate authority empty.'


        if self.user_cert != "" and self.user_cert != ".":
            status_code, message = self.check_p12_file()
            if status_code != 0:
                return status_code, message
            self.import_personal_cert()

        ## Get and import Certs
        if self.path == "":
            return 40, f'There is no path to folder of certificates. Nothing will be imported.'
        status_code, message = self.get_and_import_pem()
        if status_code != 0:
            return status_code, message
        return 0, ''



from OpenSSL import crypto
from OpenSSL.crypto import (load_privatekey, load_certificate, dump_privatekey, dump_certificate, dump_certificate_request, X509, X509Name, PKey, load_pkcs12)
from OpenSSL.crypto import (TYPE_DSA, TYPE_RSA, FILETYPE_PEM, FILETYPE_ASN1 )
from Crypto.Util import asn1

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import serialization

from pathlib import Path
import re
import hashlib

def debug_message(debug, message):
    '''
    Print debug message
    '''
    if debug == True:
        print(message)

def encrypt_string(hash_string):
    sha_signature = hashlib.sha256(hash_string.encode()).hexdigest()
    return sha_signature

def check_password_match_to_key(key_file, key_password):
    '''
    Check if password math to key and if key_file is a private key certificate
    '''
    try:
        with open(key_file, "r") as my_key_file:
            my_key_text = my_key_file.read()
            key = load_privatekey(FILETYPE_PEM, my_key_text, key_password.encode())
    except:
        return False
    return True

# source code: https://www.v13.gr/blog/?p=325
def check_cert_match_key(cert_file, key_file, key_password):
    with open(cert_file, "r") as my_cert_file:
        my_cert_text = my_cert_file.read()
        cert = load_certificate(FILETYPE_PEM, my_cert_text)
    # Read private key
    with open(key_file, "r") as my_key_file:
        my_key_text = my_key_file.read()
        priv = load_privatekey(FILETYPE_PEM, my_key_text, key_password.encode())

    pub=cert.get_pubkey()

    # Only works for RSA (I think)
    if pub.type()!=TYPE_RSA or priv.type()!=TYPE_RSA:
        raise Exception('Can only handle RSA keys')

    # This seems to work with public as well
    pub_asn1=dump_privatekey(FILETYPE_ASN1, pub)
    priv_asn1=dump_privatekey(FILETYPE_ASN1, priv)

    # Decode DER
    pub_der=asn1.DerSequence()
    pub_der.decode(pub_asn1)
    priv_der=asn1.DerSequence()
    priv_der.decode(priv_asn1)

    # Get the modulus
    pub_modulus=pub_der[1]
    priv_modulus=priv_der[1]

    if pub_modulus==priv_modulus:
        return True
    else:
        return False

def if_file_exists(file_path):
    '''
    Check if file exist
    '''
    if Path(file_path).is_file():
        return True
    else:
        return False

def if_file_is_cert(file_path):
    '''
    Check if file is pem file (certificate)
    '''
    try:
        with open(file_path, "r") as my_cert_file:
            my_cert_text = my_cert_file.read()
            cert = load_certificate(FILETYPE_PEM, my_cert_text)
    except:
        return False
    return True

def create_key_and_request(name, surname, password, cert_subject, store_location):
    '''
    Method that will create users key and request for certificate.

    It excpect that all input is valit and as input expect:
        - name (string)
        - surname (string)
        - password (string)
        - created subject for certificate (map)
        - path to location where key and certificate will be created (string)
    '''

    # Create private key
    private_key = crypto.PKey()
    private_key.generate_key( TYPE_RSA, 4096)

    # Load private key to cryptography library
    key = load_pem_private_key(crypto.dump_privatekey(crypto.FILETYPE_PEM, private_key),
                                password=None, backend=default_backend())
    # Create key with password
    password_protected_key = key.private_bytes(encoding=serialization.Encoding.PEM,
                                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                                        encryption_algorithm=serialization.BestAvailableEncryption(password.encode()))

    ## Create reuest and set subject items by cert_subject map
    csrrequest = crypto.X509Req()
    csrrequest.get_subject().CN = f'{name} {surname}'
    for (key, value) in cert_subject.items():
        setattr(csrrequest.get_subject(), key, value)
    csrrequest.set_pubkey(private_key)
    csrrequest.sign(private_key, "sha256")

    # Prepare name and path for output files
    store_location = Path(store_location)
    file_name = f'{name[0:1].lower()}{surname.split(" ")[-1].lower()}'
    key_path = store_location / f'{file_name}-key.pem'
    req_path = store_location / f'{file_name}-req.pem'
    ## Export generated key
    open(key_path, 'wb').write(password_protected_key)
    ## Export certificate request
    open(req_path, 'wb').write(crypto.dump_certificate_request(crypto.FILETYPE_PEM, csrrequest))

def create_pkcs12_cert_name(cert_file):
    file_path = Path(cert_file)
#    file_name = re.sub(file_path.suffix, '.p12', file_path.name)
    return f'{file_path.stem}.p12'

def create_pkcs12(cert_file, key_file, output_folder, key_password, p12_password):
    '''
    Method that will create users p12 certificate from user's key and certificate.

    It excpect that all input is valit and as input expect in following order:
        - path to certificate public key (string)
        - path to certificate private key(string)
        - path where p12 certificated will be created (string)
        - password for certificate private key (string)
        - password for p12 certificate (string)
    '''

    # Read certificate
    with open(cert_file, "r") as my_cert_file:
        my_cert_text = my_cert_file.read()
        cert = load_certificate(FILETYPE_PEM, my_cert_text)
    # Read private key
    with open(key_file, "r") as my_key_file:
        my_key_text = my_key_file.read()
        key = load_privatekey(FILETYPE_PEM, my_key_text, key_password.encode())

    p12 = crypto.PKCS12()
    p12.set_privatekey( key )
    p12.set_certificate( cert )

    output_file = Path(output_folder) / f'{Path(cert_file).stem}.p12'
    open(output_file, 'wb' ).write( p12.export(p12_password.encode()) )




import os
import sys
import re
import fnmatch
import subprocess, shlex
from pathlib import Path
from shutil import which
from .utils import debug_message

from OpenSSL import crypto
from OpenSSL.crypto import (load_certificate, dump_privatekey, dump_certificate, X509, X509Name, PKey, load_pkcs12)
from OpenSSL.crypto import (TYPE_DSA, TYPE_RSA, FILETYPE_PEM, FILETYPE_ASN1 )
from Crypto.Util.asn1 import (DerSequence, DerObject)

class MailClient:
    '''
    API for importing to email client
    '''
    mail_client_name = "Test mail client"
    lock = 0
    lock_message = ""
    def __init__(self, debug = False):
        print("Init")

    def import_ca(self, cert_file_path, cert_name):
        print("Import CA")
    def import_personal_cert(self, cert_file_path, password):
        print("Import personal cert")
    def import_pem(self, cert_file_path, cert_name):
        print("Import pem")

class Thunderbird:
    '''
    Class for importing certificates to Thunderbird mail client
    '''
    mail_client_name = "Thunderbird"
    def __init__(self, debug = False):
        self.thunderbird_folder= ""
        self.names_of_imported_certs = []
        self.names_of_personal_imported_certs = []
        self.modulus_of_imported_certs = []
        self.modulus_of_personal_imported_certs = []
        self.debug = debug

        self.lock, self.lock_message = self.__test_system_requirements__()
        if self.lock != 0:
            return
        self.thunderbird_folder = ""
        self.__test_existence_of_thunderbird_folder()

    def __test_system_requirements__(self):
        ## Windows check
        if os.name == "nt":
            return 1, "Program is running under operation system Windows. Import under windows is not compatible"
        ## Installed certutil program
        if which('certutil') == None or which('pk12util') == None:
            return 2, "To import certificates to Thunderbird program needs another programs called 'certutil' and 'pk12util'. They should be at 'libnss3-tools' package."
        # Unlocked
        return 0, ""
    def __test_existence_of_thunderbird_folder(self):
        # Situation if are more than one thunderbird folder
        databases = self.find('cert9.db', str(Path('~/.thunderbird/')))
        if len(databases) == 1:
            self.thunderbird_folder = databases[0].rsplit('/', 1)[0]
        elif len(databases) > 1:
            self.thunderbird_folder = databases[0].rsplit('/', 1)[0]
            debug_message(self.debug, f'There are more than one thunderbird account folders.Import will be done to the folder \'{self.thunderbird_folder}\' which is alphabetically first in the order.')
        else:
            self.lock = 3
            self.lock_message = "There is no Thunderbird folder. Install and sign to thunderbrid email client."

    ## Method finds specified file/folder
    def find(self, pattern, path):
        full_path = os.path.expanduser(path)
        result = []
        for root, dirs, files in os.walk(full_path):
            for name in files:
                if fnmatch.fnmatch(name, pattern):
                    result.append(os.path.join(root, name))
        return result

    ## Parse table from certutil command
    def process_cert_table(self, table):
        line = ''
        for i in table:
            if i == '\n':
                if line != '':
                    line = re.sub(' +', ' ', line)
                    find_flag = False
                    if len(re.findall('u,u,u$', line)) != 0:
                        find_flag = True
                    line = re.sub(' [ pPcCTuW]*,[ pPcCTuW]*,[ pPcCTuW]*$', '', line)
                    self.names_of_imported_certs.append(line)
                    if find_flag == True:
                        self.names_of_personal_imported_certs.append(line)
                line = ''
                continue
            else:
                line += i
        # Remove first and second line - header of table
        self.names_of_imported_certs.pop(0)
        self.names_of_imported_certs.pop(0)

    ## Parse modulus and exponent from certificate
    def process_modulus(self, cert):
        if not re.search('Modulus', cert) or not re.search('Exponent', cert):
            return -1, -1
        modulus = ""
        modulus = re.sub(' +', ' ', cert)
        modulus = re.sub('\n', '', modulus)
        modulus = re.sub('.*Modulus:', '', modulus)
        modulus = re.sub('Exponent:.*', '', modulus)
        modulus = re.sub(' +', '', modulus)
        modulus = re.sub(':', '', modulus)

        exponent = ""
        exponent= re.sub(' +', ' ', cert)
        exponent= re.sub('\n', '', exponent)
        exponent= re.sub('.*Exponent:', '', exponent)
        exponent= re.sub('\(.*', '', exponent)
        exponent= re.sub(' ', '', exponent)

        return modulus, exponent

    ## Source code: http://www.zedwood.com/article/python-openssl-x509-parse-certificate
    def format_split_bytes(self, aa):
        bb = aa[1:] if len(aa)%2==1 else aa #force even num bytes, remove leading 0 if necessary
        return bb

    def format_split_int(self, serial_number):
        aa = "0%x" % serial_number #add leading 0
        return self.format_split_bytes(aa)
#        return aa

    def get_modulus_and_exponent(self, x509):
        if x509.get_pubkey().type()==TYPE_RSA:
            pub_der = DerSequence()
            pub_der.decode(dump_privatekey(FILETYPE_ASN1, x509.get_pubkey()))
#            print(pub_der._seq[1])
#            modulus = "%s:%s" % ( self.format_split_int(pub_der._seq[0]), self.format_split_int(pub_der._seq[1]) )
            modulus = "%s" % ( self.format_split_int(pub_der._seq[1]) )
            exponent = pub_der._seq[2]
            return [ modulus, exponent ]
        return ''

    def get_email_address_from_cert(self, x509name, cert_path):
#        print(x509name.emailAddress)
        if x509name.emailAddress != "":
            return x509name.emailAddress
        elif x509name.CN != "":
            return x509name.CN
        else:
            return cert_path.split('/')[-1]

    ## Return specific format
    def get_org_cn_from_cert(self, x509name):
        return f'{x509name.CN} - {x509name.O}'

    def read_and_extract_cert(self, cert_file_path, ca_flag = False):
        ''' Get specific parts of certificate that are used for compare certificate is in database or not
        '''
        modulus = ""
        exponent = ""
        cert_nickname = ""

        with open(cert_file_path, 'rb+') as f:
            cert_pem = f.read()
            f.close()
            x509 = load_certificate(FILETYPE_PEM, cert_pem)
            keytype = x509.get_pubkey().type()
            keytype_list = {TYPE_RSA:'rsaEncryption', TYPE_DSA:'dsaEncryption', 408:'id-ecPublicKey'}
            key_type_str = keytype_list[keytype] if keytype in keytype_list else 'other'
            if x509.get_pubkey().type()==TYPE_RSA:
                modulus, exponent = self.get_modulus_and_exponent(x509)
#                print(modulus)

            if ca_flag == False:
                cert_nickname = self.get_email_address_from_cert(x509.get_subject(), cert_file_path)
            else:
                cert_nickname = self.get_org_cn_from_cert(x509.get_subject())
        return modulus, exponent, cert_nickname

    def cert_in_database(self, list_of_modules, modulus, exponent):
        '''
        Return if certificate is in database
        '''
        for i in list_of_modules:
            if (i[0] == modulus) and (int(i[1]) == int(exponent)):
                return True
        return False

    def get_list_of_imported_certs(self):
        '''
        Using shell command certutil for get information what certificates are at database
        '''
        if len(self.names_of_imported_certs) == 0:
#            certutil -L -d "$thunderbird_dir"
            process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-L'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            # TODO process STDERR
            self.process_cert_table(process.stdout)

    def get_modules_of_imported_certs(self):
        '''
        Get modules of imported certificates. Modules and exponent are uniq part of each certificate.
        '''
        if len(self.modulus_of_imported_certs) == 0 or len(self.modulus_of_imported_certs) != len(self.names_of_imported_certs):
           for i in self.names_of_imported_certs:
                process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-L', '-n', i], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
#                print(process.stdout)
#                print(process.returncode)
                if process.returncode == 0:
                    modulus, exponent = self.process_modulus(process.stdout)
                    if modulus != -1 and exponent != -1:
                        self.modulus_of_imported_certs.append((modulus, exponent))
                else:
                    continue

    def import_ca(self, cert_file_path, cert_name):
        '''
        Import CA to Thunderbird database.
        '''
        self.get_list_of_imported_certs()
        self.get_modules_of_imported_certs()

        modulus, exponent, cert_nickname = self.read_and_extract_cert(cert_file_path, True)
        if self.cert_in_database(self.modulus_of_imported_certs, modulus, exponent) != True:
            debug_message(self.debug, f"CA '{cert_name}' will be imported")
            #certutil -A -n $people_email -t ',,'  -d $thunderbird_dir -i $cert
            flags= "CT,C,C"
            process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-A', '-i', cert_file_path, '-t', flags, '-n', cert_nickname], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        else:
            debug_message(self.debug, f"CA '{cert_name}' is in the database")

    def import_personal_cert(self, cert_file_path, password):
        '''
        Import user's personal certificate to database. File is at pkcs12 format.
        '''
        self.get_list_of_imported_certs()

        if len(self.modulus_of_personal_imported_certs) == 0 or len(self.modulus_of_personal_imported_certs) != len(self.names_of_personal_imported_certs):
            for i in self.names_of_personal_imported_certs:
                process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-L', '-n', i], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
                if process.returncode == 0:
                    modulus, exponent = self.process_modulus(process.stdout)
                    if modulus != -1 and exponent != -1:
                        self.modulus_of_personal_imported_certs.append((modulus, exponent))
                else:
                    continue

        try:
            p12 = crypto.load_pkcs12(open(cert_file_path, 'rb').read(), password)
        except:
            # Wrong password
            return 1

        p12_cert = p12.get_certificate()
        modulus, exponent = self.get_modulus_and_exponent(p12_cert)

        if self.cert_in_database(self.modulus_of_personal_imported_certs, modulus, exponent) != True:
            #pk12util -i CA/mipytCA/userCerts/test-cert.p12 -W <password>

            debug_message(self.debug, f"'{cert_file_path}' will be imported")
            process = subprocess.run(['pk12util', '-i', cert_file_path, '-W', password,'-d', 'sql:' + self.thunderbird_folder], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        else:
            debug_message(self.debug, f"'{cert_file_path}' is in the database")


    def import_pem(self, cert_file_path, cert_name):
        '''
        Import certificate i to people tab at thunderbid
        '''
        self.get_list_of_imported_certs()
        self.get_modules_of_imported_certs()

        modulus, exponent, cert_nickname = self.read_and_extract_cert(cert_file_path)

        if self.cert_in_database(self.modulus_of_imported_certs, modulus, exponent) != True:
            debug_message(self.debug, f"'{cert_name}' will be imported")
            #certutil -A -n $people_email -t ',,'  -d $thunderbird_dir -i $cert
            flags= ",,"
            process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-A', '-i', cert_file_path, '-t', flags, '-n', cert_nickname], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        else:
            debug_message(self.debug, f"'{cert_name}' is in the database")

    def debug_delete_certs_from_path(self, cert_file_path):
        '''
        This is debugging method for easy remove from database
        '''
        cert_nickname = ""
        with open(cert_file_path, 'rb+') as f:
            cert_pem = f.read()
            f.close()
            x509 = load_certificate(FILETYPE_PEM, cert_pem)
            cert_nickname = self.get_email_address_from_cert(x509.get_subject(), cert_file_path)
        process = subprocess.run(['certutil', '-d', 'sql:' + self.thunderbird_folder, '-D', '-n', cert_nickname], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

#



from pathlib import Path
from setuptools import setup, find_packages

setup(
    name='CeMaIm',
    version='0.1',
    description='Program for import certificates to email client',
    long_description=Path('long_text.html').read_text(),
    author='Lukas Vacek',
    author_email='vaceklu6@fit.cvut.cz',
    license='GPLv3',
    url='https://github.com/vaceklu6/certs-mail-import',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    zip_safe=False,
    package_data={
        'cemaim': ['*.ui'],
    },
    install_requires=[
        'pyqt5', 'unicode', 'pycryptodome', 'pyOpenSSL', 'click', 'requests', 'bs4', 'unidecode', 'importlib.resources'
    ],
    entry_points={
        'console_scripts': [
            'cemaim = cemaim:cli',
        ],
    },
)



import configparser
import click
import getpass
from .cemaim import *
from .utils import create_pkcs12
from .cemaim_qt import Gui
from PyQt5 import QtWidgets, uic, QtCore, QtGui

def parse_cert_path(ctx, param, value):
    return value

def parse_auth_cert(ctx, param, value):
    return value

def validate_config(ctx, param, value):
    '''
    Validate if config File exist
    '''
    if value == None: return value
    try:
        config = configparser.ConfigParser()
        with open(value) as f:
            config.read_file(f)
    except Exception:
        raise click.BadParameter('config file can not be readed')
    return value

def parse_config_file(config_file):
    '''
    Read config file and fill variables to map
    '''
    config = configparser.ConfigParser()
    config.optionxform = str
    with open(config_file) as f:
        config.read_file(f)

    config_map = {}

    for i in config.sections():
        section = i[0:3]
        for key in config[i]:
            config_map[f'{section}_{key.lower()}'] = config[i][key]

    return config_map

def force_parameters_over_config_file(config_map, folder, ca, mail, user_cert, password):
    '''
    Overwrite variables from config file with parametrs settings
    '''
    if ca!= None:
        config_map['imp_ca'] = ca

    if folder != None:
        config_map['imp_folder'] = folder

    if ('imp_mail' not in config_map.keys()) and (mail == None):
        config_map['imp_mail'] = "thb"
    if ('imp_mail' not in config_map.keys()) and (mail != None):
        config_map['imp_mail'] = mail
    if ('imp_mail' in config_map.keys()):
        if config_map['imp_mail'] not in mail_clients.keys():
            print(f'There is no mail client called {config_map["imp_mail"]}. Client it set to default client \'thb\'')
            config_map['imp_mail'] = "thb"

    if user_cert != None:
        config_map['imp_user_cert'] = user_cert

    config_map['password'] = password



@click.command('cemaim')
@click.option('-F', '--folder', callback=parse_cert_path,
        help='Path to location where personal certificates are stored.')
@click.option('-C', '--ca', callback=parse_auth_cert,
        help='Specification of certification authority that will be imported to mail client.')
@click.option('-c', '--cli', is_flag=True,
        help='Do not open window and it will be runned from terminal. This option is not config file option.')
@click.option('-d', '--debug', is_flag=True,
        help='Debug mode is compatible only with \'--cli\' option. This option is not config file option.')
@click.option('-f', '--config-file', metavar='FILENAME', type=click.Path(), callback=validate_config,
        help='Read configuration from config file.')
@click.option('-m', '--mail', metavar=cli_mail_client_text()[0],
type=click.Choice(mail_clients.keys()),
        help='Choose mail client where will be certificates imported. By default will be choosed Thunderbird')
@click.option('-p', '--password', is_flag=True,
        help='Write password for user certificate.')
@click.option('-u', '--user-cert', metavar='FILENAME', type=click.Path(),
        help='Specification of user\'s certificate.')
def cli(folder, ca, cli, debug, config_file, mail, password, user_cert):
    """
    Program for import users certificates to mail client

    It also run Gui with parameters from command line
    """

    config_map = {}
    config_map['imp_ca'] = ""
    config_map['imp_folder'] = ""
    config_map['imp_user_cert'] = ""
    password_string=""
    if password == True:
        try:
            password_string = getpass.getpass()
        except Exception as error:
            print('ERROR', error)


    if config_file != None:
        config_map = parse_config_file(config_file)

    force_parameters_over_config_file(config_map, folder, ca, mail, user_cert, password_string)

    if cli == True:
        ## Import certificate from command line
        cemaim = CeMaIm(config_map['imp_ca'], config_map['imp_folder'], config_map['imp_user_cert'], config_map['imp_mail'], password_string, debug)
        status_code, message = cemaim.run()
        if status_code != 0:
           print(message, file=sys.stderr)
           sys.exit(status_code)

    else:
        ## Run Gui
        gui = Gui(config_map)
        gui.run()


def main():
    cli()
